@@ -54,7 +54,8 @@ void hd_scan_sysfs_block(hd_data_t *hd_data)
 
   if(hd_probe_feature(hd_data, pr_block_mods)) {
     PROGRESS(1, 0, "block modules");
-    load_module(hd_data, "ide_cd");
+    load_module(hd_data, "ide-cd");
+    load_module(hd_data, "ide-disk");
     load_module(hd_data, "sr_mod");
     load_module(hd_data, "sd_mod");
 #if !defined(__s390__) && !defined(__s390x__)
@@ -162,7 +163,10 @@ void get_block_devs(hd_data_t *hd_data)
     if(bus_id) bus_id++;
 
     bus_name = NULL;
-    if((s = hd_read_sysfs_link(sf_dev, "bus"))) {
+    if(
+      (s = hd_read_sysfs_link(sf_dev, "subsystem")) ||
+      (s = hd_read_sysfs_link(sf_dev, "bus"))
+    ) {
       bus_name = strrchr(s, '/');
       if(bus_name) bus_name++;
       bus_name = new_str(bus_name);
