@@ -49,6 +49,11 @@ static unsigned char pci_cfg_byte(pci_t *pci, int fd, unsigned idx);
 static void dump_pci_data(hd_data_t *hd_data);
 static void hd_read_macio(hd_data_t *hd_data);
 static void hd_read_vio(hd_data_t *hd_data);
+static void hd_read_xen(hd_data_t *hd_data);
+static void hd_read_ps3_system_bus(hd_data_t *hd_data);
+static void hd_read_platform(hd_data_t *hd_data);
+static void add_xen_network(hd_data_t *hd_data);
+static void add_xen_storage(hd_data_t *hd_data);
 
 void hd_scan_sysfs_pci(hd_data_t *hd_data)
 {
@@ -75,9 +80,19 @@ void hd_scan_sysfs_pci(hd_data_t *hd_data)
 
   hd_read_macio(hd_data);
 
-  PROGRESS(3, 0, "vio");
+  PROGRESS(4, 0, "vio");
 
   hd_read_vio(hd_data);
+
+  PROGRESS(5, 0, "xen");
+
+  hd_read_xen(hd_data);
+  PROGRESS(6, 0, "ps3");
+
+  hd_read_ps3_system_bus(hd_data);
+  
+  PROGRESS(7, 0, "platform");
+  hd_read_platform(hd_data);
 }
 
 
@@ -543,7 +558,7 @@ void dump_pci_data(hd_data_t *hd_data)
 void hd_read_macio(hd_data_t *hd_data)
 {
   char *s, *t;
-  char *macio_name, *macio_type, *macio_compat;
+  char *macio_name, *macio_type, *macio_compat, *macio_modalias;
   hd_t *hd, *hd2;
   str_list_t *sf_bus, *sf_bus_e;
   char *sf_dev;
@@ -564,7 +579,7 @@ void hd_read_macio(hd_data_t *hd_data)
       hd_sysfs_id(sf_dev)
     );
 
-    macio_name = macio_type = macio_compat = NULL;
+    macio_name = macio_type = macio_compat = macio_modalias = NULL;
 
     if((s = get_sysfs_attr_by_path(sf_dev, "name"))) {
       macio_name = canon_str(s, strlen(s));
@@ -581,6 +596,11 @@ void hd_read_macio(hd_data_t *hd_data)
       ADD2LOG("    compatible = \"%s\"\n", macio_compat);
     }
 
+    if((s = get_sysfs_attr_by_path(sf_dev, "modalias"))) {
+      macio_modalias = canon_str(s, strlen(s));
+      ADD2LOG("    modalias = \"%s\"\n", macio_modalias);
+    }
+
     if(
       macio_type && (
         !strcmp(macio_type, "network") ||
@@ -608,6 +628,9 @@ void hd_read_macio(hd_data_t *hd_data)
       s = hd_sysfs_find_driver(hd_data, hd->sysfs_id, 1);
       if(s) add_str_list(&hd->drivers, s);
 
+      hd->modalias = macio_modalias;
+      macio_modalias = NULL;
+
       s = new_str(hd->sysfs_id);
 
       if((t = strrchr(s, '/'))) {
@@ -712,5 +735,282 @@ void hd_read_vio(hd_data_t *hd_data)
   free_str_list(sf_bus);
 }
 
+/*
+ * Get platform data from sysfs.
+ */
+void hd_read_platform(hd_data_t *hd_data)
+{
+  char *s, *platform_type;
+  int scsi_cnt = 0;
+  hd_t *hd;
+  str_list_t *sf_bus, *sf_bus_e;
+  char *sf_dev;
+
+  sf_bus = reverse_str_list(read_dir("/sys/bus/platform/devices", 'l'));
+
+  if(!sf_bus) {
+    ADD2LOG("sysfs: no such bus: platform\n");
+    return;
+  }
+
+  for(sf_bus_e = sf_bus; sf_bus_e; sf_bus_e = sf_bus_e->next) {
+    sf_dev = new_str(hd_read_sysfs_link("/sys/bus/platform/devices", sf_bus_e->str));
+
+    ADD2LOG(
+      "  platform device: name = %s\n    path = %s\n",
+      sf_bus_e->str,
+      hd_sysfs_id(sf_dev)
+    );
+
+    platform_type = NULL;
+
+    if((s = get_sysfs_attr_by_path(sf_dev, "modalias"))) {
+      platform_type = canon_str(s, strlen(s));
+      ADD2LOG("    type = \"%s\"\n", platform_type);
+    }
+
+    if(
+      platform_type && !strcmp(platform_type, "ps3_storage")
+    ) {
+      hd = add_hd_entry(hd_data, __LINE__, 0);
+      hd->bus.id = bus_ps3_system_bus;
+
+      hd->vendor.id = MAKE_ID(TAG_PCI, 0x104d); /* Sony */
+
+      hd->base_class.id = bc_storage;
+      hd->sub_class.id = sc_sto_scsi;
+      hd->slot = scsi_cnt++;
+      hd->device.id = MAKE_ID(TAG_SPECIAL, 0x1000); /* PS3_DEV_TYPE_STOR_DISK */
+      str_printf(&hd->device.name, 0, "PS3 storage %d", hd->slot);
+
+      hd->modalias = new_str(platform_type);
+
+      hd->sysfs_id = new_str(hd_sysfs_id(sf_dev));
+      hd->sysfs_bus_id = new_str(sf_bus_e->str);
+      s = hd_sysfs_find_driver(hd_data, hd->sysfs_id, 1);
+      if(s) add_str_list(&hd->drivers, s);
+    }
+
+    platform_type = free_mem(platform_type);
+
+    free_mem(sf_dev);
+  }
+
+  free_str_list(sf_bus);
+}
+
+
+
+/*
+ * Get ps3 data from sysfs.
+ */
+static void hd_read_ps3_system_bus(hd_data_t *hd_data)
+{
+  char *s, *ps3_name;
+  int eth_cnt = 0; 
+  hd_t *hd;
+  str_list_t *sf_bus, *sf_bus_e;
+  char *sf_dev;
+
+  sf_bus = reverse_str_list(read_dir("/sys/bus/ps3_system_bus/devices", 'l'));
+
+  if(!sf_bus) {
+    ADD2LOG("sysfs: no such bus: ps3_system_bus\n");
+    return;
+  }
+
+  for(sf_bus_e = sf_bus; sf_bus_e; sf_bus_e = sf_bus_e->next) {
+    sf_dev = new_str(hd_read_sysfs_link("/sys/bus/ps3_system_bus/devices", sf_bus_e->str));
+
+    ADD2LOG(
+      "  ps3 device: name = %s\n    path = %s\n",
+      sf_bus_e->str,
+      hd_sysfs_id(sf_dev)
+    );
+
+    ps3_name = NULL;
+
+    if((s = get_sysfs_attr_by_path(sf_dev, "modalias"))) {
+      ps3_name = canon_str(s, strlen(s));
+      ADD2LOG("    modalias = \"%s\"\n", ps3_name);
+    }
+
+    if(
+      ps3_name && (
+        !strcmp(ps3_name, "ps3:3") ||
+        !strcmp(ps3_name, "gelic_net")
+      )
+    ) {
+      hd = add_hd_entry(hd_data, __LINE__, 0);
+      hd->bus.id = bus_ps3_system_bus;
+
+      hd->vendor.id = MAKE_ID(TAG_PCI, 0x104d); /* Sony */
+
+      hd->base_class.id = bc_network;
+      hd->sub_class.id = 0;	/* ethernet */
+      hd->slot = eth_cnt++;
+      hd->device.id = MAKE_ID(TAG_SPECIAL, 0x1003); /* PS3_DEV_TYPE_SB_GELIC */
+      str_printf(&hd->device.name, 0, "PS3 Ethernet card %d", hd->slot);
+
+      hd->modalias = new_str(ps3_name);
+
+      hd->sysfs_id = new_str(hd_sysfs_id(sf_dev));
+      hd->sysfs_bus_id = new_str(sf_bus_e->str);
+      s = hd_sysfs_find_driver(hd_data, hd->sysfs_id, 1);
+      if(s) add_str_list(&hd->drivers, s);
+    }
+
+    ps3_name = free_mem(ps3_name);
+
+    free_mem(sf_dev);
+  }
+
+  free_str_list(sf_bus);
+}
+
+/*
+ * Get xen (network & storage) data from sysfs.
+ */
+void hd_read_xen(hd_data_t *hd_data)
+{
+  char *s, *xen_type, *xen_node;
+  int eth_cnt = 0, blk_cnt = 0;
+  hd_t *hd;
+  str_list_t *sf_bus, *sf_bus_e;
+  char *sf_dev, *drv, *module;
+  unsigned u;
+
+  sf_bus = reverse_str_list(read_dir("/sys/bus/xen/devices", 'l'));
+
+  if(!sf_bus) {
+    ADD2LOG("sysfs: no such bus: xen\n");
+
+    if(hd_is_xen(hd_data)) {
+      add_xen_network(hd_data);
+      add_xen_storage(hd_data);
+    }
+
+    return;
+  }
+
+  for(sf_bus_e = sf_bus; sf_bus_e; sf_bus_e = sf_bus_e->next) {
+    sf_dev = new_str(hd_read_sysfs_link("/sys/bus/xen/devices", sf_bus_e->str));
+
+    ADD2LOG(
+      "  xen device: name = %s\n    path = %s\n",
+      sf_bus_e->str,
+      hd_sysfs_id(sf_dev)
+    );
+
+    xen_type = xen_node = NULL;
+
+    if((s = get_sysfs_attr_by_path(sf_dev, "devtype"))) {
+      xen_type = canon_str(s, strlen(s));
+      ADD2LOG("    type = \"%s\"\n", xen_type);
+    }
+
+    if((s = get_sysfs_attr_by_path(sf_dev, "nodename"))) {
+      xen_node = canon_str(s, strlen(s));
+      ADD2LOG("    node = \"%s\"\n", xen_node);
+    }
+
+    drv = new_str(hd_read_sysfs_link(sf_dev, "driver"));
+
+    s = new_str(hd_read_sysfs_link(drv, "module"));
+    module = new_str(s ? strrchr(s, '/') + 1 : NULL);
+    free_mem(s);
+
+    ADD2LOG("    module = \"%s\"\n", module);
+
+    if(
+      xen_type &&
+      (
+        !strcmp(xen_type, "vif") ||
+        !strcmp(xen_type, "vbd")
+      )
+    ) {
+      hd = add_hd_entry(hd_data, __LINE__, 0);
+      hd->bus.id = bus_none;
+
+      hd->vendor.id = MAKE_ID(TAG_SPECIAL, 0x6011);	/* xen */
+
+      if(!strcmp(xen_type, "vif")) {	/* network */
+        hd->base_class.id = bc_network;
+        hd->sub_class.id = 0;	/* ethernet */
+        hd->slot = eth_cnt++;
+        u = 3;
+        if(module) {
+          if(!strcmp(module, "xennet")) u = 1;
+          if(!strcmp(module, "xen_vnif")) u = 2;
+        }
+        hd->device.id = MAKE_ID(TAG_SPECIAL, u);
+        str_printf(&hd->device.name, 0, "Virtual Ethernet Card %d", hd->slot);
+      }
+      else {	/* storage */
+        hd->base_class.id = bc_storage;
+        hd->sub_class.id = sc_sto_other;
+        hd->slot = blk_cnt++;
+        u = 3;
+        if(module) {
+          if(!strcmp(module, "xenblk")) u = 1;
+          if(!strcmp(module, "xen_vbd")) u = 2;
+        }
+        hd->device.id = MAKE_ID(TAG_SPECIAL, 0x1000 + u);
+        str_printf(&hd->device.name, 0, "Virtual Storage %d", hd->slot);
+      }
+
+      hd->rom_id = new_str(xen_node);
+
+      hd->sysfs_id = new_str(hd_sysfs_id(sf_dev));
+      hd->sysfs_bus_id = new_str(sf_bus_e->str);
+      s = hd_sysfs_find_driver(hd_data, hd->sysfs_id, 1);
+      if(s) add_str_list(&hd->drivers, s);
+    }
+
+    free_mem(sf_dev);
+    free_mem(drv);
+    free_mem(module);
+  }
+
+  free_str_list(sf_bus);
+
+  /* maybe only one of xen_vnif, xen_vbd was loaded */
+  if(!eth_cnt && !hd_module_is_active(hd_data, "xen_vnif")) add_xen_network(hd_data);
+  if(!blk_cnt && !hd_module_is_active(hd_data, "xen_vbd")) add_xen_storage(hd_data);
+}
+
+
+/*
+ * fake xen network device
+ */
+void add_xen_network(hd_data_t *hd_data)
+{
+  hd_t *hd;
+
+  hd = add_hd_entry(hd_data, __LINE__, 0);
+  hd->base_class.id = bc_network;
+  hd->sub_class.id = 0;	/* ethernet */
+  hd->vendor.id = MAKE_ID(TAG_SPECIAL, 0x6011);	/* xen */
+  hd->device.id = MAKE_ID(TAG_SPECIAL, 0x0002);	/* xen-vnif */
+  hd->device.name = new_str("Virtual Ethernet Card");
+}
+
+
+/*
+ * fake xen storage controller
+ */
+void add_xen_storage(hd_data_t *hd_data)
+{
+  hd_t *hd;
+
+  hd = add_hd_entry(hd_data, __LINE__, 0);
+  hd->base_class.id = bc_storage;
+  hd->sub_class.id = sc_sto_other;
+  hd->vendor.id = MAKE_ID(TAG_SPECIAL, 0x6011);	/* xen */
+  hd->device.id = MAKE_ID(TAG_SPECIAL, 0x1002);	/* xen-vbd */
+  hd->device.name = new_str("Virtual Storage");
+}
+
+
 /** @} */
 
